@@ -1,42 +1,37 @@
-/*-
-# X-BASED ABACUS
-#
-#  AbacusC.c
-#
-###
-#
-#  Copyright (c) 1992 - 2005	David Albert Bagley, bagleyd@tux.org
-#
-#                   All Rights Reserved
-#
-#  Permission to use, copy, modify, and distribute this software and
-#  its documentation for any purpose and without fee is hereby granted,
-#  provided that the above copyright notice appear in all copies and
-#  that both that copyright notice and this permission notice appear in
-#  supporting documentation, and that the name of the author not be
-#  used in advertising or publicity pertaining to distribution of the
-#  software without specific, written prior permission.
-#
-#  This program is distributed in the hope that it will be "useful",
-#  but WITHOUT ANY WARRANTY; without even the implied warranty of
-#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-#
-#  Taken from a C++ group project where I was a lead contributer
-#  OOP Group4!
-*/
+/*
+ * @(#)AbacusC.c
+ *
+ * Copyright 1992 - 2017  David A. Bagley, bagleyd AT verizon.net
+ *
+ * All rights reserved.
+ *
+ * Permission to use, copy, modify, and distribute this software and
+ * its documentation for any purpose and without fee is hereby granted,
+ * provided that the above copyright notice appear in all copies and
+ * that both that copyright notice and this permission notice appear in
+ * supporting documentation, and that the name of the author not be
+ * used in advertising or publicity pertaining to distribution of the
+ * software without specific, written prior permission.
+ *
+ * This program is distributed in the hope that it will be "useful",
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+ *
+ * Taken from a C++ group project where I was a lead contributer
+ * OOP Group4!
+ */
 
 /* Methods string infix calculator file for Abacus */
-
 #include "AbacusP.h"
 #ifndef WINVER
+extern int signgam;
+#endif
 #include <math.h>
 #ifndef M_PI
 #define M_PI 3.14159265358979323846
 #endif
 #ifndef M_E
 #define M_E 2.7182818284590452354
-extern int signgam;
-#endif
 #endif
 
 /* proposed, implemented only a small portion
@@ -52,11 +47,11 @@ $	summation
 %	mod e.g. 7%2=1 [medium]
 &	bitwise and [medium]
 '
-(	(
+(	( [equate]
 )	) [equate]
 *	multipication [medium]
 +	addition [low]
-,	real imaginary delimitor [numeric]
+,	[numeric]
 -	subtraction (not sign change) [low]
 .	decimal point (or octal pt if in octal, etc.) [numeric]
 /	divide e.g. 1/2=0.5 [medium]
@@ -83,42 +78,42 @@ C	hexadecimal 12 [numeric]
 D	hexadecimal 13 [numeric]
 E	hexadecimal 14 [numeric]
 F	hexadecimal 15 [numeric]
-G	variance (sigma^2)
-H	hyperbolic
-I	inverse
-J
+G	16 [numeric]
+H	17 [numeric]
+I	18 [numeric]
+J	19 [numeric]
 K
 L	logarithm base 2
 M	memory recall ('M', ('0'-'9' | 'A'-'F'))
 N	natural logarithm
-O       bitwise xor [high]
+O	bitwise xor [high]
 P	permutation [low]
 Q	quit [quit]
 R	"to the root of" e.g. 8R3=2 [high]
 S	sample variance (s^2)
-T       square summation
-U       x^3
+T	square summation
+U	x^3
 V	x^2
-W       convert from MinSec
+W	convert from MinSec
 X	e^x (unfortunately 'e' & 'E' are used already)
-Y
-Z
+Y	hyperbolic
+Z	inverse
 [
 \	Pascal's div i.e 28\8=3 [medium]
 ]
 ^	"power of" e.g. 2^3=8 [high]
-_       reset statistics
+_	reset statistics
 `
-a       clear almost all [clear]
+a	clear almost all [clear]
 b	base mode ('b', (['2'-'9'] | '1',['0'-'6']))
 c	combination [low]
-d       decimal hot key (also 'b', '1', '0')
+d	decimal hot key (also 'b', '1', '0')
 e	used for exponents e.g. 6.02*10^22 = 6.02e22(input) = 6.02 22(output)
 f
 g	gradient mode
-h       hexadecimal hot key (also 'b', '1', '6')
+h	hexadecimal hot key (also 'b', '1', '6')
 i	invert 1/x
-j
+j	variance (sigma^2)
 k	cosine (kosine)
 l	logarithm base 10
 m	memory ('m', ('0'-'9' | 'A'-'F'))
@@ -131,8 +126,8 @@ s	sine
 t	tangent
 u	cube root
 v	unary operation square root (v-) UNIX's dc & bc use this also
-w       convert to MinSec
-x       complex mode (base 10 only)
+w	convert to MinSec
+x	complex mode (base 10 only)
 y
 z	clear (zero) [clear]
 {
@@ -164,16 +159,16 @@ static const short unsigned int orderASCII[] =
 
   constant,  notused,  notused,  notused, /* @ A B C */
    notused,  notused,  notused, constant, /* D E F G */
-      mode,     mode,  notused,  notused, /* H I J K */
+   notused,  notused,  notused,  notused, /* H I J K */
      unary, constant,    unary,     high, /* L M N O */
        low,  notused,     high, constant, /* P Q R S */
   constant,    unary,    unary,    unary, /* T U V W */
-     unary,  notused,  notused,  notused, /* X Y Z [ */
+     unary,     mode,     mode,  notused, /* X Y Z [ */
     medium,  notused,     high,     mode, /* \ ] ^ _ */
 
    notused,  notused,     mode,      low, /* ` a b c */
       mode, constant,  notused,     mode, /* d e f g */
-      mode,    unary,  notused,    unary, /* h i j k */
+      mode,    unary, constant,    unary, /* h i j k */
      unary,     mode,    unary,     mode, /* l m n o */
   constant,  notused,     mode,    unary, /* p q r s */
      unary,    unary,    unary,    unary, /* t u v w */
@@ -203,8 +198,8 @@ static const short unsigned int operationASCII[] =
   operation, operation, operation, operation, /* < = > ? */
 
   operation,   numeric,   numeric,   numeric, /* @ A B C */
-    numeric,   numeric,   numeric, operation, /* D E F G */
-  operation, operation, undefined, undefined, /* H I J K */
+    numeric,   numeric,   numeric,   numeric, /* D E F G */
+    numeric,   numeric,   numeric, undefined, /* H I J K */
   operation, operation, operation, operation, /* L M N O */
   operation,      quit, operation, operation, /* P Q R S */
   operation, operation, operation, operation, /* T U V W */
@@ -213,7 +208,7 @@ static const short unsigned int operationASCII[] =
 
   undefined,     clear, operation, operation, /* ` a b c */
   operation, operation, undefined, operation, /* d e f g */
-  operation, operation, undefined, operation, /* h i j k */
+  operation, operation, operation, operation, /* h i j k */
   operation, operation, operation, operation, /* l m n o */
   operation,     clear, operation, operation, /* p q r s */
   operation, operation, operation, operation, /* t u v w */
@@ -223,35 +218,43 @@ static const short unsigned int operationASCII[] =
 
 #define CHAR_TO_DIGIT(c) ((c >= 'A') ? c - 'A' + 10 : c - '0')
 #define DIGIT_TO_CHAR(d) ((d >= 10) ? (char) ('A' + d - 10) : (char) ('0' + d))
-#define IS_DIGIT(c) ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'F'))
+#define IS_DIGIT(c) ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'J'))
+#define IS_NUMBER(c) (IS_DIGIT(c) || c == decimalPoint)
 /* True for period, +/-, or operator */
 #define IS_VALID(v, b) ((IS_DIGIT(v)) ? (CHAR_TO_DIGIT(v) < b) : True)
-#define MAXVALUELENGTH 64
+#define MAX_VALUE_LENGTH 64
+#define CALC_STRING_SIZE 256
 
-enum StackUsage {paren, order};
+enum Stackusage {paren, order};
 
 typedef struct _Term
 {
 	double variable;
 	char operation;
-	Boolean orderUsage;
+	Boolean orderusage;
 	struct _Term *previous;
 } Term;
 
 typedef struct _Expression
 {
 	/* char *variable; */
-	char variable[MAXVALUELENGTH]; /* Hard code a limit for now */
+	char variable[MAX_VALUE_LENGTH]; /* Hard code a limit for now */
 	char operation;
 	struct _Expression *previous;
 } Expression;
 
 /* struct Stack */
 static Term *term;
+#ifdef EXTRA
 static Expression *expression;
+#endif
 
-#define DEFAULTBASE 10
-#define DEFAULTFRACTDIGITS 16
+#if 0
+#define DEBUG 1
+#endif
+#define DEFAULT_BASE 10
+#define DEFAULT_HALFBASE 5
+#define DEFAULT_DECIMAL_DIGITS 16
 
 /* Inputter & Parser & Paren Manager */
 static Boolean period = False;
@@ -260,20 +263,34 @@ static Boolean negateNext = False; /* handles unary '-' */
 static double left, right; /* sides of a binary operation */
 static char pendingOperation; /* operation to be performed on left (& right if not unary) */
 static Boolean hub; /* intermediate result? usually a ')' or pi pressed */
-static char memoryBuf[64];
-static char displayBuf[64];
+static char memoryBuf[CALC_STRING_SIZE];
+static char displayBuf[CALC_STRING_SIZE];
+static char stringBuf[CALC_STRING_SIZE];
 static int digits = 1;
-static int currentBase = DEFAULTBASE, fractDigits = DEFAULTFRACTDIGITS;
+static int currentBase = DEFAULT_BASE, currentDigits = DEFAULT_DECIMAL_DIGITS;
+static char decimalPoint = '.';
+static char groupSymbol = ',';
 static int nestingLevel = 0;
+#ifdef EXTRA
 static Expression parseExpression;
+#endif
 
+#ifdef WINVER
+/* Some versions have, some do not, change to 1 if you do not */
 #if 0
+static double
+pow(double x, double y)
+{
+	return exp(y * log(x));
+}
+
 static double
 cbrt(double x) /* -inf < x < inf */
 {
 	return ((x < 0.0) ? -pow(-x, 1.0 / 3.0) : pow(x, 1.0 / 3.0));
 }
 #endif
+#endif
 
 static double
 powInt(int base, int y)
@@ -292,150 +309,196 @@ powInt(int base, int y)
 	return z;
 }
 
-/* More exact (int) (log(x) / log ((double) base)) */
+/* More exact (int) (log(x) / log(base)) */
 static int
 logInt(double x, int base)
 {
+	double quotient = x;
 	int i = 0;
 
-	if (x >= (double) base)
-		while (x >= (double) base) {
-			x /= (double) base;
+	if (quotient >= base) {
+		while (quotient >= base) {
+			quotient /= base;
 			i++;
 		}
-	else if (x > 0.0)
-		while (x < 1.0) {
-			x *= (double) base;
+		if (quotient + 0.0000000001 >= base)
+			i++;
+	} else if (quotient > 0.0) {
+		while (quotient < 1.0) {
+			quotient *= base;
 			i--;
 		}
+	}
 	return i;
 }
 
-static double
-rootInt(double x, int y) /* y != 0 && (x >= 0 || (y odd)) */
+#ifdef EXTRA
+double
+rootInt(double x, int y)
 {
-	if (x < 0.0 && ((double) y / 2.0) != ((double) (y / 2)))
-		return -pow (-x, 1.0 / (double) y);
-	else if (x == 0.0)
+	/* y != 0 && (x >= 0 || (y odd)) */
+	if (x < 0.0 && 2 * (y / 2) != y)
+		return -pow(-x, 1.0 / y);
+	else if (x <= 0.0)
 		return 0.0;
 	else
-		return pow (x, 1.0 / (double) y);
+		return pow(x, 1.0 / y);
 }
+#endif
 
-static double
-convertToDecimal(int b, char *inputstring)
+double
+convertToDecimal(int base, char *inputString)
 {
 	int k = 0;
 	Boolean negative = False;
-	double number = 0.0;
 	int digit;
 	int length = 0;
+	double number = 0.0;
 	double factor;
 
 	/* Convert Integer Part */
 	k = 0;
-	if (inputstring[k] == '-')
-	{
+	if (inputString[k] == '-') {
 		negative = True;
 		k++;
 	}
-	while (IS_DIGIT(inputstring[k + length]))
+	while (IS_DIGIT(inputString[k + length]))
 		length++;
-	factor = powInt(b, length);
-	for (; length > 0; length--, k++)
-	{
-		digit = CHAR_TO_DIGIT(inputstring[k]);
-		factor /= ((double) b);
-		number += ((double) digit) * factor;
+	factor = powInt(base, length);
+	for (; length > 0; length--, k++) {
+		digit = CHAR_TO_DIGIT(inputString[k]);
+		factor /= base;
+		number += factor * digit;
 	}
 
 	/* Convert Fractional Part */
-	if (inputstring[k] == '.')
-	{
+	if (inputString[k] == decimalPoint) {
 		k++;
-		while (IS_DIGIT(inputstring[k]))
-		{
-			digit = CHAR_TO_DIGIT(inputstring[k]);
-			factor /= ((double) b);
-			number += ((double) digit) * factor;
+		while (IS_DIGIT(inputString[k])) {
+			digit = CHAR_TO_DIGIT(inputString[k]);
+			factor /= base;
+			number += factor * digit;
 			k++;
 		}
 	}
 
-	if ((Boolean) negative)
+	if (negative)
 		number = -number;
 	negative = False;
 
 	return number;
 }
 
-static void
-convertFromDecimal(char *outputString, int b, double x)
+void
+convertFromInteger(char *outputString, int base, int x)
 {
-	char string[MAXVALUELENGTH];
-	Boolean localPeriod = False;
-	double number = 0.0;
-	int places_base;
+	char string[MAX_VALUE_LENGTH];
+	int number = x;
+	int placesBase;
 	int l = 0, i = 0;
 	short unsigned int digit;
 	double factor;
-	int fractdigits = fractDigits;
+	int fractDigits = currentDigits;
 
-	(void) sprintf(string, "%g", x);
+	(void) sprintf(string, "%d", number);
 	if (string[i] == '-') {
 		outputString[l++] = '-';
-		x = -x;
+		number = -number;
 	}
 	while (string[i] != '\0') {
-		if (string[i] == '.')
+		i++;
+	}
+	{
+		/* Chicken and egg problem:
+		   rounding might increase placesBase */
+		placesBase = logInt(number, base);
+		fractDigits -= (placesBase + 1);
+		/* rounder */
+		number += (int) (pow((double) base, -fractDigits) / 2.0);
+
+		/* Convert Integer Part */
+		if (number < 1.0) {
+			outputString[l++] = '0';
+			factor = 1.0 / base;
+		} else {
+			placesBase = logInt(number, base);
+			factor = pow((double) base, (double) placesBase);
+			for (; placesBase >= 0; placesBase--) {
+				digit = (short unsigned int) (number / factor);
+				outputString[l++] = DIGIT_TO_CHAR(digit);
+				number -= (int) (factor * digit);
+				factor /= base;
+			}
+		}
+	}
+	outputString[l] = '\0';
+}
+
+void
+convertFromDouble(char *outputString, int base, double x)
+{
+	char string[MAX_VALUE_LENGTH];
+	Boolean localPeriod = False;
+	double number = x;
+	int placesBase;
+	int l = 0, i = 0;
+	short unsigned int digit;
+	double factor;
+	int fractDigits = currentDigits;
+
+	(void) sprintf(string, "%g", number);
+	if (string[i] == '-') {
+		outputString[l++] = '-';
+		number = -number;
+	}
+	while (string[i] != '\0') {
+		if (string[i] == '.') /* DECIMAL_POINT LOCALE C */
 			localPeriod = True;
 		i++;
 	}
 	{
 		/* Chicken and egg problem:
-		   rounding might increase places_base */
-		places_base = logInt(x, b);
-		fractdigits -= (places_base + 1);
+		   rounding might increase placesBase */
+		placesBase = logInt(number, base);
+		fractDigits -= (placesBase + 1);
 		/* rounder */
-		x += (pow((double) b, (double) -fractdigits) / 2.0);
+		number += (pow((double) base, -fractDigits) / 2.0);
 
-		number = x;
 		/* Convert Integer Part */
 		if (number < 1.0) {
 			outputString[l++] = '0';
-			factor = 1.0 / ((double) b);
+			factor = 1.0 / base;
 		} else {
-			places_base = logInt(number, b);
-			factor = pow((double) b, (double) places_base);
-			for (; places_base >= 0; places_base--) {
-				digit = (int) (number / factor);
+			placesBase = logInt(number, base);
+			factor = pow((double) base, (double) placesBase);
+			for (; placesBase >= 0; placesBase--) {
+				digit = (short unsigned int) (number / factor);
 				outputString[l++] = DIGIT_TO_CHAR(digit);
-				number -= digit * factor;
-				factor /= ((double) b);
+				number -= factor * digit;
+				factor /= base;
 			}
 		}
 		/* Convert Fractional Part */
 		if (localPeriod) {
-			outputString[l++] = '.';
-			for (places_base = 1; places_base <= fractdigits;
-					places_base++) {
-				digit = (int) (number / factor);
+			outputString[l++] = decimalPoint;
+			for (placesBase = 1; placesBase <= fractDigits;
+					placesBase++) {
+				digit = (short unsigned int) (number / factor);
 				outputString[l++] = DIGIT_TO_CHAR(digit);
-				number -= digit * factor;
-				factor /= ((double) b);
+				number -= factor * digit;
+				factor /= base;
 			}
 			while (outputString[l - 1] == '0')
 				l--;
 		}
 	}
-
 	outputString[l] = '\0';
 }
 
 static double
 formatFromDisplay(int base, char *string)
 {
-	char floatString[MAXVALUELENGTH];
+	char floatString[MAX_VALUE_LENGTH];
 	Boolean got1Digit = False;
 	short unsigned int periods = 0;
 	int s = 0, k = 0;
@@ -444,11 +507,7 @@ formatFromDisplay(int base, char *string)
 		if (IS_DIGIT(string[s])) {
 			got1Digit = True;
 			floatString[k++] = string[s];
-		} else switch (string[s]) {
-		case '-':
-			floatString[k++] = '-';
-			break;
-		case '.':
+		} else if (string[s] == decimalPoint) {
 			if (periods == 1)
 				(void) printf("'%c' handler not implemented in Format_From_Display\n", string[s]);
 			periods++;
@@ -456,9 +515,10 @@ formatFromDisplay(int base, char *string)
 				got1Digit = True;
 				floatString[k++] ='0';
 			}
-			floatString[k++] = '.';
-			break;
-		default:
+			floatString[k++] = decimalPoint;
+		} else if (string[s] == '-') {
+			floatString[k++] = '-';
+		} else {
 			(void) printf("'%c' handler not implemented in Format_From_Display\n", string[s]);
 		}
 		s++;
@@ -472,13 +532,13 @@ formatFromDisplay(int base, char *string)
 }
 
 static void
-formatToDisplay (char *string, double z)
+formatToDisplay (char *string, double z, int base)
 {
 	int i = 0, j, length;
 	short unsigned int periods = 0;
 
-	convertFromDecimal(string, currentBase, z);
-	length = strlen(string);
+	convertFromDouble(string, base, z);
+	length = (int) strlen(string);
 	while (i <= length) {
 		if (string[i] == '\0' && periods == 0) {
 			j = length;
@@ -487,10 +547,10 @@ formatToDisplay (char *string, double z)
 				j--;
 			}
 			length++;
-			string[i++] = '.';
+			string[i++] = decimalPoint;
 			periods++;
 		}
-		if (string[i] == '.')
+		if (string[i] == decimalPoint)
 			periods++;
 		if (string[i] == '+') {
 			j = i;
@@ -519,13 +579,26 @@ static void reset(void)
 static void setBase(int base)
 {
 	currentBase = base;
-	fractDigits = logInt(powInt(DEFAULTBASE, DEFAULTFRACTDIGITS), base);
+	currentDigits = logInt(powInt(DEFAULT_BASE, DEFAULT_DECIMAL_DIGITS), base);
 }
 
+static void setDecimalComma(Boolean dpc)
+{
+	if (dpc) {
+		decimalPoint = ',';
+		groupSymbol = '.';
+	} else {
+		decimalPoint = '.';
+		groupSymbol = ',';
+	}
+}
+
+#ifdef EXTRA
 static int getNestingLevel(void)
 {
 	return nestingLevel;
 }
+#endif
 
 static void incNestingLevel(void)
 {
@@ -570,7 +643,7 @@ evaluateSingle(double arg, char unaryOperation)
 			return 0.0;
 		}
 #else
-		if ((arg < 0.0 && ((double)((int) arg) == arg)) ||
+		if ((arg < 0.0 && ((int) arg) == arg) ||
 				arg > 20.0) {
 			/* this does not catch all the errors */
 			return 0.0;
@@ -628,17 +701,19 @@ evaluateDouble(double arg1, char binaryOperation, double arg2)
 	return comp;
 }
 
+#ifdef EXTRA
 static Boolean
-initStack(void)
+newStack(void)
 {
 	Term *plate;
 
 	if (!(plate = (Term *) malloc(sizeof(Term))))
 		return False;
 	plate->previous = NULL;
-	term = plate->previous;
+	term = plate;
 	return True;
 }
+#endif
 
 static Boolean
 emptyStack(void)
@@ -657,12 +732,14 @@ flushStack(void)
 	}
 }
 
+#ifdef EXTRA
 static void
-delStack(void)
+deleteStack(void)
 {
 	flushStack();
 	/* free(term); */
 }
+#endif
 
 static Boolean
 pushStack(double z, char c, Boolean u)
@@ -673,7 +750,7 @@ pushStack(double z, char c, Boolean u)
 		return False;
 	plate->variable = z;
 	plate->operation = c;
-	plate->orderUsage = u;
+	plate->orderusage = u;
 	plate->previous = term;
 	term = plate;
 	return True;
@@ -685,7 +762,7 @@ static void popStack(double *z, char *c, Boolean *u)
 
 	*z = term->variable;
 	*c = term->operation;
-	*u = term->orderUsage;
+	*u = term->orderusage;
 	term = plate->previous;
 	free(plate);
 }
@@ -695,9 +772,9 @@ static char topOp(void)
 	return term->operation;
 }
 
-static Boolean topUsage(void)
+static Boolean topusage(void)
 {
-	return term->orderUsage;
+	return term->orderusage;
 }
 
 static Boolean emptyExpression(void)
@@ -707,13 +784,13 @@ static Boolean emptyExpression(void)
 
 static Boolean previousOrder(void)
 {
-	return ((Boolean) (!emptyExpression() && topUsage() == order));
+	return (!emptyExpression() && topusage() == order);
 }
 
 static Boolean canReduceExpression(char binaryOperation)
 {
-	return ((Boolean) (previousOrder() && orderASCII[(int) topOp()] >=
-			orderASCII[(int) binaryOperation]));
+	return (previousOrder() && orderASCII[(int) topOp()] >=
+		orderASCII[(int) binaryOperation]);
 }
 
 static void resetWholeExpression(void)
@@ -741,16 +818,18 @@ static Boolean storeExpressionParen(double *myLeft, char *binaryOperation)
 static void evaluateExpressionPart(double *myLeft, char binaryOperation,
 		double myRight)
 {
-	if (binaryOperation != '\0')
-		*myLeft = evaluateDouble(*myLeft, binaryOperation, myRight);
-	else
+	if (binaryOperation == '\0')
 		*myLeft = myRight;
+	else
+		*myLeft = evaluateDouble(*myLeft, binaryOperation, myRight);
 }
 
 static Boolean evaluateExpressionOrder(double *myLeft, char *binaryOperation,
 	double *myRight, char newOperation)
 {
-	if (*binaryOperation != '\0') {
+	if (*binaryOperation == '\0') {
+		*myLeft = *myRight;
+	} else {
 		if (orderASCII[(int) *binaryOperation] >=
 				orderASCII[(int) newOperation]) {
 			*myLeft = evaluateDouble(*myLeft, *binaryOperation,
@@ -770,8 +849,7 @@ static Boolean evaluateExpressionOrder(double *myLeft, char *binaryOperation,
 				return False;
 			}
 		}
-	} else
-		*myLeft = *myRight;
+	}
 	return True;
 }
 
@@ -779,7 +857,7 @@ static void evaluateExpressionParen(double *myLeft, char *binaryOperation)
 {
 	double myRight;
 
-	while (!emptyExpression() && topUsage() == order) {
+	while (!emptyExpression() && topusage() == order) {
 		myRight = *myLeft;
 		getPreviousExpressionPart(myLeft, binaryOperation);
 		evaluateExpressionPart(myLeft, *binaryOperation, myRight);
@@ -793,15 +871,16 @@ static void evaluateExpression(double *myLeft, char *binaryOperation)
 	while (!emptyExpression()) {
 		myRight = *myLeft;
 		getPreviousExpressionPart(myLeft, binaryOperation);
-		if (binaryOperation != NULL)
+		if (binaryOperation == NULL)
+			*myLeft = myRight;
+		else
 			*myLeft = evaluateDouble(*myLeft, *binaryOperation,
 				myRight);
-		else
-			*myLeft = myRight;
 	} /* ignore uneven "(()" */
 	*binaryOperation = '\0';
 }
 
+#ifdef EXTRA
 static void
 initParser(void)
 {
@@ -818,6 +897,7 @@ resetRight(void)
 	right = 0.0;
 	hub = True;
 }
+#endif
 
 static void
 resetExpression(void)
@@ -829,8 +909,10 @@ resetExpression(void)
 }
 
 static void
-parse(double input, Boolean numeral, char newOperation)
+parse(double input, Boolean gotNumeral, char newOperation)
 {
+	Boolean numeral = gotNumeral;
+
 	if (numeral) {
 		if (negateNext) {
 			negateNext = False;
@@ -902,9 +984,15 @@ parse(double input, Boolean numeral, char newOperation)
 		if (pendingOperation == '\0')
 			left = right;
 		hub = True;
-	} else if (orderASCII[(int) newOperation] != notused) {
+	} else if (orderASCII[(int) newOperation] == notused) {
+		resetExpression();
+	} else {
 		/* binary operation */
-		if (!numeral) {
+		if (numeral) {
+			(void) evaluateExpressionOrder(&left,
+				&pendingOperation, &right, newOperation);
+			right = left;
+		} else {
 			if (previousOrder()) /* this stops 2*+ = 6 (2*2+2) */ {
 				getPreviousExpressionPart(&left,
 					&pendingOperation);
@@ -917,26 +1005,20 @@ parse(double input, Boolean numeral, char newOperation)
 			} else if (newOperation == '-') {
 				negateNext = True;
 			}
-		} else {
-			(void) evaluateExpressionOrder(&left,
-				&pendingOperation, &right, newOperation);
-			right = left;
 		}
 		pendingOperation = newOperation;
-	} else { /* not used operation */
-		resetExpression();
 	}
-	formatToDisplay(memoryBuf, right);
+	formatToDisplay(memoryBuf, right, currentBase);
 	(void) strcpy(displayBuf, memoryBuf); /* update display */
 }
 
 static void
 inputOperator(char input)
 {
-	if (!gotFirstDigit) {
-		parse(0.0, False, input);
-	} else {
+	if (gotFirstDigit) {
 		parse(formatFromDisplay(currentBase, memoryBuf), True, input);
+	} else {
+		parse(0.0, False, input);
 	}
 	reset();
 }
@@ -959,17 +1041,15 @@ inputNumeric(char input)
 			digits++;
 			memoryBuf[digits] = '\0';
 		}
-	} else switch(input) {
-	case '.':
+	} else if (input == decimalPoint) {
 		if (period) {
-			reset(); /* Handle multiple '.' */
+			reset(); /* Handle multiple decimalPoint */
 		}
 		period = True;
 		memoryBuf[digits] = input;
 		digits++;
 		memoryBuf[digits] = '\0';
-		break;
-	case '~': /* Handle +/- */
+	} else if (input == '~') { /* Handle +/- */
 		/* Note: if `memoryBuf == "0"' then "+/-" should act
 		   as a unary operation, that is, if no number is entered,
 		   it should act on the result. */
@@ -980,7 +1060,7 @@ inputNumeric(char input)
 		if (memoryBuf[0] == '0' && memoryBuf[1] == '\0')
 			return False;
 		while (memoryBuf[i] != ' ' && memoryBuf[i] != '-' &&
-			memoryBuf[i] != ',' && i != 0)
+			memoryBuf[i] != groupSymbol && i != 0)
 			i--;
 		if (memoryBuf[i] == '-') {
 			/* Take a '-' from front of a negative number */
@@ -991,7 +1071,7 @@ inputNumeric(char input)
 			digits--;
 		} else {
 			/* Put the '-' in the front of a positive number */
-			if (digits == MAXVALUELENGTH - 1)
+			if (digits == MAX_VALUE_LENGTH - 1)
 				return False;
 			j = digits;
 			while (j >= i) {
@@ -1001,74 +1081,186 @@ inputNumeric(char input)
 			memoryBuf[i] = '-';
 			digits++;
 		}
-		break;
-	default:
+	} else {
 		reset();
 		return False;
 	}
 	return True;
 }
 
-static void
-convertStringToAbacus(AbacusWidget w, char *string)
+void
+convertStringToAbacus(AbacusWidget w, const char *string, int aux)
 {
 	int decimal = 0;
-	int num, factor, i, sign = 0, len = strlen(string);
+	int num, factor, i, sign = 0, len = (int) strlen(string);
+	AbacusWidget *wap = &w, wa;
 
-	ClearRails(w);
-	while (string[decimal] != '\0' && string[decimal] != '.')
+	if (aux == 1 && w->abacus.leftAux != NULL)
+		wap = (AbacusWidget *) w->abacus.leftAux;
+	else if (aux == 2 && w->abacus.rightAux != NULL)
+		wap = (AbacusWidget *) w->abacus.rightAux;
+	wa = *wap;
+	if (aux == 0) {
+		clearRails(w);
+#ifndef WINVER
+	} else {
+		clearAuxRails(w, aux);
+#endif
+	}
+	while (string[decimal] != '\0' && string[decimal] != decimalPoint)
 		decimal++;
 	if (string[0] == '-' || string[0] == '+') {
 		sign = 1;
 	}
 	for (i = 0; i < decimal - sign; i++) {
-		/* w->abacus.displayBase == w->abacus.base and all that ... */
+		/* wa->abacus.displayBase == wa->abacus.base and all that ... */
 		num = char2Int(string[decimal - 1 - i]);
-		factor = num / w->abacus.decks[TOP].factor;
-		if (factor > 0) {
-			SetAbacusMove(w, ABACUS_MOVE, 0, 1, i, factor);
-			num = num - factor * w->abacus.decks[TOP].factor;
+		if (wa->abacus.decks[BOTTOM].number <= wa->abacus.base / 2) {
+			factor = num / wa->abacus.decks[TOP].factor;
+			if (factor > 0) {
+				setAbacusMove(w, ACTION_MOVE, aux, 1, i, factor);
+				num = num - factor * wa->abacus.decks[TOP].factor;
+			}
 		}
-		factor = num / w->abacus.decks[BOTTOM].factor;
+		factor = num / wa->abacus.decks[BOTTOM].factor;
 		if (factor > 0) {
-			SetAbacusMove(w, ABACUS_MOVE, 0, 0, i, factor);
+			setAbacusMove(w, ACTION_MOVE, aux, 0, i, factor);
 		}
 	}
-	if (w->abacus.sign && string[0] == '-') {
-		SetAbacusMove(w, ABACUS_MOVE, 0, 0, w->abacus.rails -
-			w->abacus.decimalPosition - 1, 1);
+	if (wa->abacus.sign && string[0] == '-') {
+		setAbacusMove(w, ACTION_MOVE, aux, 0, wa->abacus.rails -
+			wa->abacus.decimalPosition - 1, 1);
 	}
 	for (i = 0; i < len - decimal - 1 &&
-			i < w->abacus.decimalPosition; i++) {
+			i < wa->abacus.decimalPosition; i++) {
+		int offset = 0;
+
 		num = char2Int(string[decimal + i + 1]);
-		factor = num / w->abacus.decks[TOP].factor;
-		if (factor > 0) {
-			SetAbacusMove(w, ABACUS_MOVE, 0, 1, -i - 1, factor);
-			num = num - factor * w->abacus.decks[TOP].factor;
+		if  (wa->abacus.decks[BOTTOM].piece != 0)
+			offset++;
+		if  (wa->abacus.decks[BOTTOM].piecePercent != 0 &&
+				i >= wa->abacus.shiftPercent)
+			offset++;
+		if (wa->abacus.decks[BOTTOM].number <= wa->abacus.base / 2) {
+			factor = num / wa->abacus.decks[TOP].factor;
+			if (factor > 0) {
+				setAbacusMove(w, ACTION_MOVE, aux, 1, -i - 1 - offset, factor);
+				num = num - factor * wa->abacus.decks[TOP].factor;
+			}
 		}
-		factor = num / w->abacus.decks[BOTTOM].factor;
+		factor = num / wa->abacus.decks[BOTTOM].factor;
 		if (factor > 0) {
-			SetAbacusMove(w, ABACUS_MOVE, 0, 0, -i - 1, factor);
+			setAbacusMove(w, ACTION_MOVE, aux, 0, -i - 1 - offset, factor);
 		}
 	}
 }
 
-void calculate(AbacusWidget w, char * buffer)
+void
+addBackAnomaly(char * buf, int anomaly, int shift, int base)
+{
+	double anom;
+	int factor;
+	int shiftValue = (int) multPower(1, base, shift);
+	int anomalyValue = anomaly * shiftValue / base;
+
+	anom = convertToDecimal(base, buf);
+	factor = (int) (anom / (shiftValue - anomalyValue));
+	anom = anom + factor * anomalyValue;
+	convertFromDouble(buf, base, anom);
+}
+
+void
+zeroFractionalPart(char * buf)
 {
 	unsigned int i;
 
-	if (w->abacus.displayBase != w->abacus.base)
-		return;
-	currentBase = w->abacus.displayBase;
-	fractDigits = logInt(powInt(DEFAULTBASE, DEFAULTFRACTDIGITS),
-		currentBase);
-	resetExpression();
+	for (i = 0; i < strlen(buf); i++) {
+		if (buf[i] == '.') {
+			buf[i + 1] = '\0';
+			return;
+		}
+	}
+}
+
+char *strAddChar(char *buf, char c)
+{
+	char *s;
+
+	s = buf;
+	s = s + strlen(buf); /* the position to write */
+	*s = c; /* this should be on the place where the null char first was. */
+	s++;
+	*s = '\0'; /* add a new null string */
+	return buf;
+}
+
+char *strAddInt(char *buf, int value)
+{
+	char *s;
+
+	s = buf;
+	s = s + strlen(buf); /* the position to write */
+	(void) sprintf(s, "%d", value); /* always base 10 */
+	return buf;
+}
+
+static void
+convertStringBases(char *sb, char *buf, int displayBase, int base) {
+	int i;
+
+	for (i = 0; buf[i] != '\0'; i++) {
+		if (IS_NUMBER(buf[i])) {
+			char  numberString[CALC_STRING_SIZE];
+			double num;
+			int j = 0;
+
+			while (buf[i] != '\0' && IS_NUMBER(buf[i])) {
+				numberString[j] = buf[i];
+				i++;
+				j++;
+			}
+			i--;
+			numberString[j] = '\0';
+			num = convertToDecimal(displayBase,
+				numberString);
+			convertFromDouble(numberString, base, num);
+			(void) strcat(sb, numberString);
+#ifdef DEBUG
+			(void) printf("ns =%s\n", numberString);
+#endif
+		} else {
+			sb = strAddChar(sb, buf[i]);
+		}
+	}
+}
+
+void calculate(AbacusWidget w, char * buffer, int aux)
+{
+	unsigned int i;
+
+#ifdef DEBUG
+	(void) printf("buffer =%s\n", buffer);
+#endif
 	memoryBuf[0] = '0';
 	memoryBuf[1] = '\0';
 	displayBuf[0] = '0';
 	displayBuf[1] = '\0';
-	for (i = 0; i < strlen(buffer); i++) {
-		switch (operationASCII[(int) buffer[i]]) {
+	stringBuf[0] = '0';
+	stringBuf[1] = '\0';
+	setBase(w->abacus.base);
+	setDecimalComma(w->abacus.decimalComma);
+	if (w->abacus.displayBase != w->abacus.base) {
+		convertStringBases(stringBuf, buffer, w->abacus.displayBase,
+			w->abacus.base);
+#ifdef DEBUG
+		(void) printf("conv buffer =%s\n", stringBuf);
+#endif
+	} else {
+		(void) strcpy(stringBuf, buffer);
+	}
+	resetExpression();
+	for (i = 0; i < strlen(stringBuf); i++) {
+		switch (operationASCII[(int) stringBuf[i]]) {
 		case undefined:
 #ifdef DEBUG
 			(void) printf("undefined\n");
@@ -1081,15 +1273,19 @@ void calculate(AbacusWidget w, char * buffer)
 			break;
 		case numeric:
 #ifdef DEBUG
-			(void) printf("number %c %s\n", buffer[i], memoryBuf);
+			if (stringBuf[i] == groupSymbol)
+				(void) printf("ignore\n");
+			else
+				(void) printf("number %c %s\n", stringBuf[i], memoryBuf);
 #endif
-			(void) inputNumeric(buffer[i]);
+			if (stringBuf[i] != groupSymbol)
+				(void) inputNumeric(stringBuf[i]);
 			break;
 		case operation:
 #ifdef DEBUG
-			(void) printf("operate %c %s\n", buffer[i], memoryBuf);
+			(void) printf("operate %c %s\n", stringBuf[i], memoryBuf);
 #endif
-			inputOperator(buffer[i]);
+			inputOperator(stringBuf[i]);
 			break;
 		case clear:
 #ifdef DEBUG
@@ -1105,7 +1301,20 @@ void calculate(AbacusWidget w, char * buffer)
 	}
 	inputOperator('=');
 #ifdef DEBUG
-	(void) printf("displayBuf %s\n", displayBuf);
+	(void) printf("display =%s\n", displayBuf);
 #endif
-	convertStringToAbacus(w, displayBuf);
+	if (w->abacus.anomaly != 0) {
+		addBackAnomaly(displayBuf, w->abacus.anomaly,
+			w->abacus.shiftAnomaly,
+			w->abacus.base);
+	}
+	if (w->abacus.anomalySq != 0) {
+		addBackAnomaly(displayBuf, w->abacus.anomalySq,
+			w->abacus.shiftAnomaly + w->abacus.shiftAnomalySq,
+			w->abacus.base);
+	}
+	if (w->abacus.subdeck != 0) {
+		zeroFractionalPart(displayBuf);
+	}
+	convertStringToAbacus(w, displayBuf, aux);
 }
